"""Registry module that contains singletons."""

from ._registry import get_bg_async, get_exchange, get_monitor, get_renderer, get_trace_handler, set_renderer

__all__ = [
    "get_bg_async",
    "get_exchange",
    "get_monitor",
    "get_renderer",
    "get_trace_handler",
    "set_renderer",
]
